'''
File: GUI.py
Description: load all class for medImgProc
             Contains externally usable class
History:
    Date    Programmer SAR# - Description
    ---------- ---------- ----------------------------
  Author: w.x.chan@gmail.com         12JAN2019           - Created
Author: w.x.chan@gmail.com           08OCT2019           - v1.4.0
                                                              -added colortoggler
Author: w.x.chan@gmail.com           08OCT2019           - v1.5.2
                                                              -added Intensity scaling
                                                              - lower slider
Author: w.x.chan@gmail.com           10Jan2020           - v2.3.9
                                                              -added cubic spline line drawing
                                                              -removed latex dependency
                                                              -debug function show() in image2DGUI
Author: w.x.chan@gmail.com           12Jan2020           - v2.3.10
                                                              -debug keypress switch frame of rgb image
Author: w.x.chan@gmail.com           23MAR2020           - v2.6.4
                                                              -added color contour 
Author: w.x.chan@gmail.com           24MAR2020           - v2.6.13
                                                              -allow save image
Requirements:
    numpy.py
    matplotlib.py
    imageio.py

Known Bug:
    HSV color format not supported
All rights reserved.
'''
_version='2.6.13'
import logging
logger = logging.getLogger(__name__)
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from matplotlib.widgets import Slider
from matplotlib.widgets import Button
from scipy import interpolate
#matplotlib.rc('text', usetex=True)
try:
    import tkinter
    from tkinter import filedialog
    import os
except:
    pass
'''
variables
'''
STD_INSTRUCTIONS='Press Enter to save and exit, Esc to exit\n '
'''
Internal Use functions
'''
def getLastTwoDimArray(imageArray,dimLoad,color=0):
    outputArray=np.copy(imageArray)
    if len(outputArray.shape)>(2+color):
        outputArray=getLastTwoDimArray(outputArray[dimLoad[0]],dimLoad[1:],color=color)
    return outputArray
def getFramePts(pts,dimLoad):
    newPts=np.array(pts)
    for n in range(len(dimLoad)-2):
        if len(newPts)!=0:
            filterFrame=(newPts[:,n]==dimLoad[n])
            newPts=newPts[filterFrame,:]
        else:
            break
    return newPts
def dimToTitle(dimension,showIndex):
    titleOutput=''
    for n in range(-1,-len(dimension)-1,-1):
        titleOutput+=dimension[n]+':'+str(showIndex[n])
        if n==-1:
            titleOutput+='^v   '  #r'$\updownarrow$'
            
        else:
            titleOutput+='<>'#r'$\leftrightarrow$ \hspace{1cm} '
            break
    #titleOutput+='\n '+r'$\leftarrow$ $\rightarrow$ \hspace{1cm} $\uparrow$ $\downarrow$'
    
    return titleOutput

'''
Main GUI class
'''
class image2DGUI:
    def __init__(self,imageClass,addInstruct='',disable=[],initPointList=None,initLineList=None,manualToleranceRatio=0.05,showNow=True,contourImageArray=None):
        self.title=None
        self.addInstruct=STD_INSTRUCTIONS
        if addInstruct!='':
            self.addInstruct+=addInstruct+'\n '
        if type(imageClass)==str:
            self.image=medImgProc.imread(imageClass)
        self.image=imageClass.clone()
        self.image.data=np.maximum(0,self.image.data)
        if contourImageArray is None:
            self.contourImage=None
        else:
            self.contourImage=imageClass.clone()
            self.contourImage.data[:]=contourImageArray
            self.contourImage.data=self.contourImage.data.astype(int)
        self.color=0
        if 'RGB' in self.image.dim:
            self.image.rearrangeDim('RGB',False)
            if self.contourImage is not None:
                self.contourImage.rearrangeDim('RGB',False)
            self.color=1
        elif 'RGBA' in self.image.dim:
            self.image.rearrangeDim('RGBA',False)
            if self.contourImage is not None:
                self.contourImage.rearrangeDim('RGBA',False)
            self.color=1
        if self.color==1:
            self.addInstruct+='press 1,2,3,.. to toggler color channel and 0 to show all.\n '
            self.colorToggler=[]
        self.fig=plt.figure(1)
        self.showIndex=[]
        for n in range(len(self.image.data.shape)-self.color):
            self.showIndex.append(0)
        self.disable=disable
        self.sliderLoc=0.02
        self.connectionID=[]
        if 'click' not in self.disable or 'line' not in self.disable:
            self.connectionID.append(self.fig.canvas.mpl_connect('button_press_event', self.onclick))
        if 'line' not in self.disable:
            self.sliderLoc+=0.06
            self.connectionID.append(self.fig.canvas.mpl_connect('button_release_event', self.onrelease))
            self.connectionID.append(self.fig.canvas.mpl_connect('motion_notify_event', self.onmotion))
            
        self.connectionID.append(self.fig.canvas.mpl_connect('key_press_event',self.onKeypress))   
        self.enter=False
        self.scaleVisual=1.
        self.logVisual=0.
        self.line_selected=-1
        self.point_selected=-1
        self.show_line=False
        '''
        Return parameters
        '''
        if initPointList is None:
            self.points=np.empty((0,len(self.image.dim)))
        else:
            '''sanitize to show'''
            initPointList=np.copy(initPointList)
            for n in range(len(initPointList)):
                for m in range(len(initPointList[n])-2):
                    initPointList[n,m]=np.round(initPointList[n,m])
            self.points=initPointList
            
        
        self.lineplt=[]
        self.lines=[]
        if initLineList is not None:
            for line in initLineList:
                self.lines.append(line.copy())
                
        self.axslide=[]
        self.sSlide=[]
        
        self.manualToleranceRatio=manualToleranceRatio
        self.manualTolerance=(max(*self.image.data.shape[-2:])*self.manualToleranceRatio)**2.
        
        self.loadImageFrame()
        if showNow:
            plt.show()
    def sliderUpdate(self,val):
        self.line_selected=-1
        for n in range(len(self.showIndex)-2):
            self.showIndex[n]=int(self.sSlide[n].val)
        self.scaleVisual=self.sSlide[-2].val
        self.logVisual=self.sSlide[-1].val
        self.showNewFrame()
        if self.show_line:
            for nline in range(len(self.lines)):
                self.showNewLine(nline)
    def getLineIndex(self,y,x):
        chosen=None
        distance=float('inf')
        if self.line_selected<0:
            for nline in range(len(self.lines)):
                points=getFramePts(self.lines[nline],self.showIndex)
                if len(points)>1:
                    tck,temp = interpolate.splprep([points[:,-1], points[:,-2]], s=0,k=min(4,len(points))-1)
                    cspline_detectline = np.array(interpolate.splev(np.arange(0, 1.01, 0.01), tck)).T
                    distance_temp=np.min(np.sum((cspline_detectline-np.array([[x,y]]))**2.,axis=1))
                    if distance_temp<distance:
                        distance=distance_temp
                        chosen=nline
                elif len(points)==1:
                    distance_temp=np.min(np.sum((np.array([points[:,-1], points[:,-2]])-np.array([[x,y]]))**2.,axis=1))
                    if distance_temp<distance:
                        distance=distance_temp
                        chosen=nline
        elif len(self.lines[self.line_selected])>0:
            points=getFramePts(self.lines[self.line_selected],self.showIndex)
            distance=np.sum((points[:,-2:]-np.array([[y,x]]))**2.,axis=1)
            chosen=np.argmin(distance)
        return (chosen,distance)
    def onclick(self,event):
        if not(event.dblclick) and event.button==1 and event.inaxes==self.ax:
            newPt=np.array([*self.showIndex[:-2],event.ydata,event.xdata])
            if 'click' not in self.disable:
                self.points=np.vstack((self.points,newPt))
                self.showNewPoints()
            if self.show_line:
                lineIndex,distance=self.getLineIndex(event.ydata,event.xdata)
                if self.line_selected>=0:
                    addlinept=len(self.lines[self.line_selected])
                    if lineIndex is not None:
                        if distance.min() < self.manualTolerance:
                            self.point_selected=lineIndex
                            addlinept=None
                        elif addlinept>1:
                            detectline=getFramePts(self.lines[self.line_selected],self.showIndex)
                            tck,temp = interpolate.splprep([detectline[:,-1], detectline[:,-2]], s=0,k=min(4,len(detectline))-1)
                            cspline_detectline = np.array(interpolate.splev(np.arange(0, 1.+1./100./(addlinept-1), 1./100./(addlinept-1)), tck)).T
                            detectdistance=np.sum((cspline_detectline-np.array([[event.xdata,event.ydata]]))**2.,axis=1)
                            if detectdistance.min()<self.manualTolerance:
                                splineIndex=detectdistance.argmin()
                                for npoint in range(1,addlinept):
                                    if splineIndex<np.argmin(np.sum((cspline_detectline-np.array([[self.lines[self.line_selected][npoint][-1],self.lines[self.line_selected][npoint][-2]]]))**2.,axis=1)):
                                        addlinept=npoint
                                        break
                    if addlinept is not None:
                        self.point_selected=-1
                        self.lines[self.line_selected].insert(addlinept,newPt)
                    self.showNewLine(self.line_selected)
                elif lineIndex is not None:
                    if self.line_selected==-1:
                        self.line_selected=lineIndex
                        self.showNewLine(lineIndex)
                    elif self.line_selected==-2:
                        self.lines.pop(lineIndex)
                        self.line_selected=-1
                        self.loadImageFrame()
                    
    def onrelease(self,event):
        self.point_selected=-1
        if self.line_selected>=0:
            if len(self.lines[self.line_selected])>0:
                self.showNewLine(self.line_selected)
        return
    def onmotion(self,event):
        if self.line_selected<0 or self.point_selected<0:
            return
        if event.xdata is None or event.ydata is None:
            if self.line_selected>=0 and self.point_selected>=0:
                self.lines[self.line_selected].pop(self.point_selected)
                self.point_selected=-1
                if len(self.lines[self.line_selected])>0:
                   self.showNewLine(self.line_selected)
                else:
                   self.lines.pop(self.line_selected)
                   self.line_selected=-1
                   self.loadImageFrame()
            return
        newPt=np.array([*self.showIndex[:-2],event.ydata,event.xdata])
        self.lines[self.line_selected][self.point_selected]=newPt.copy()
        self.showNewLine(self.line_selected)
        return
    def onKeypress(self,event):
        if event.key == 'escape':#return values and quit
            for connection in self.connectionID:
                self.fig.canvas.mpl_disconnect(connection)
            plt.close(event.canvas.figure)
        elif event.key == 'enter':
            self.enter=True
            for connection in self.connectionID:
                self.fig.canvas.mpl_disconnect(connection)
            plt.close(event.canvas.figure)
        elif event.key=='up':
            self.switchFrame(-3,1)
        elif event.key=='down':
            self.switchFrame(-3,-1)
        elif event.key=='right':
            self.switchFrame(-4,1)
        elif event.key=='left':
            self.switchFrame(-4,-1)
        elif event.key=='ctrl+z':
            self.removeLastPoint()
        elif self.color==1 and event.key=='0':
            self.colorToggler=[]
            self.showNewFrame()
        elif self.color==1 and event.key in ['1','2','3','4','5','6','7','8','9']:
            if (int(event.key)-1) in self.colorToggler:
                self.colorToggler.remove(int(event.key)-1)
            elif self.image.data.shape[-1] > (int(event.key)-1):
                self.colorToggler+=[int(event.key)-1]
            self.showNewFrame()
        elif event.key in self.image.dim:
            self.swapFrame(event.key)
        else:
            logger.info(event.key)
    def save_image(self,event):
        self.ax.axis('off')
        extent = self.ax.get_window_extent().transformed(self.fig.dpi_scale_trans.inverted())
        maxdpi=max(self.image.data.shape[-2-self.color]/extent.height,self.image.data.shape[-1-self.color]/extent.width)
        root = tkinter.Tk()
        fileName=root.withdraw()
        fileName=''.join(self.image.dim[(-2-self.color):][:2])
        for dimN in range(len(self.image.dim)-2-self.color):
            fileName+='_'+self.image.dim[dimN]+str(self.showIndex[dimN])
        tempdir = filedialog.asksaveasfilename(parent=root, initialdir=os.getcwd(), title='Save Image '+fileName+' as')
        if isinstance(tempdir,str):
            if tempdir[-4:]!='.png':
                tempdir+='.png'
            self.fig.savefig(tempdir, bbox_inches=extent,dpi=maxdpi)
        root.destroy()
        self.ax.axis('on')
    def togger_line(self,event):
        self.show_line=not(self.show_line)
        if not(self.show_line):
            self.line_selected=-1
        self.loadImageFrame()
    def new_line(self,event):
        self.show_line=True
        if len(self.lineplt)>self.line_selected and self.line_selected>=0:
            if len(self.lineplt[self.line_selected])>1:
                self.lineplt[self.line_selected][1].set_color('b')
        self.line_selected=len(self.lines)
        self.lines.append([])
    def edit_line(self,event):
        if len(self.lineplt)>self.line_selected:
            if len(self.lineplt[self.line_selected])>1:
                self.lineplt[self.line_selected][1].set_color('b')
        self.line_selected=-1
    def del_line(self,event):
        self.line_selected=-2
    def switchFrame(self,index,val=1):
        self.line_selected=-1
        if len(self.showIndex)>=(-index):
            self.showIndex[index]+=val
            if self.showIndex[index]>(self.image.data.shape[len(self.showIndex)+index]-1):
                self.showIndex[index]=0
            elif self.showIndex[index]<0:
                self.showIndex[index]=self.image.data.shape[len(self.showIndex)+index]-1
            self.sSlide[len(self.showIndex)+index].set_val(self.showIndex[index])
        if self.show_line:
            self.loadImageFrame()
        else:
            self.showNewFrame()
    def swapFrame(self,axis):
        self.line_selected=-1
        if 'swap' not in self.disable:
            if self.color:
                transposeIndex=self.image.rearrangeDim([axis,self.image.dim[-1]],arrangeFront=False)
                if self.contourImage is not None:
                    self.contourImage.rearrangeDim([axis,self.image.dim[-1]],arrangeFront=False)
            else:
                transposeIndex=self.image.rearrangeDim([axis],arrangeFront=False)
                if self.contourImage is not None:
                    self.contourImage.rearrangeDim([axis],arrangeFront=False)
            newShowIndex=[]
            if self.color:
                transposeIndex=transposeIndex[:-1]
            for n in range(len(transposeIndex)):
                newShowIndex.append(self.showIndex[transposeIndex[n]])
            self.showIndex=newShowIndex
            self.points=self.points[:,transposeIndex]
            for nline in range(len(self.lines)):
                self.lines[nline]=self.lines[nline][:,transposeIndex]
            self.manualTolerance=(max(*self.image.data.shape[-2:])*self.manualToleranceRatio)**2.
            self.loadImageFrame()
            self.showNewFrame()
    def showNewFrame(self):
        newShowImage=getLastTwoDimArray(self.image.data,self.showIndex,color=self.color)
        if self.color:
            newShowImage[...,tuple(self.colorToggler)]=0
        newShowImage=np.maximum(0,np.minimum(255,(newShowImage*self.scaleVisual)**(10.**self.logVisual))).astype('uint8')
        self.main.set_data(newShowImage)
        self.ax.set_aspect(self.image.dimlen[self.image.dim[-2-self.color]]/self.image.dimlen[self.image.dim[-1-self.color]])
        if self.contourImage is not None:
            if self.contour is not None:
                for coll in self.contour.collections:
                    coll.remove()
            showContour=getLastTwoDimArray(self.contourImage.data,self.showIndex,color=0)
            getlevels=np.arange(0.5+showContour.min(),showContour.max(),1)
            if len(getlevels)>0:
                self.contour=self.ax.contour(showContour,getlevels,linewidths=1.2)
            else:
                self.contour=None
        self.showNewPoints()
        
        pp=plt.setp(self.title,text=self.addInstruct+dimToTitle(self.image.dim[:-2-self.color],self.showIndex[:-2]))
        self.fig.canvas.draw()
    def showNewPoints(self):
        showpoints=getFramePts(self.points,self.showIndex)
        self.ptplt.set_offsets(showpoints[:,[-1,-2]])
        self.fig.canvas.draw()
    def showNewLine(self,lineIndex):
        temp_color='b'
        if lineIndex==self.line_selected:
            temp_color='r'
        showline=getFramePts(self.lines[lineIndex],self.showIndex)
        if len(showline)<=0:
            self.lineplt[lineIndex][0].set_visible(False)
            self.lineplt[lineIndex][1].set_visible(False)
        else:
            if len(self.lineplt)<=lineIndex:
                self.loadImageFrame()
                return;
            elif type(self.lineplt[lineIndex])!=list:
                self.loadImageFrame()
                return;
            self.lineplt[lineIndex][0].set_offsets(showline[:,[-1,-2]])
            self.lineplt[lineIndex][0].set_visible(True)
            self.lineplt[lineIndex][1].set_visible(True)
            if len(showline)>1:
                self.lineplt[lineIndex][0].set_color('b')
                tck,temp = interpolate.splprep([showline[:,-1], showline[:,-2]], s=0,k=min(4,len(showline))-1)
                cspline_line = interpolate.splev(np.arange(0, 1.01, 0.01), tck)
                self.lineplt[lineIndex][1].set_xdata(cspline_line[0])
                self.lineplt[lineIndex][1].set_ydata(cspline_line[1])
                self.lineplt[lineIndex][1].set_color(temp_color)
                #self.lineplt[lineIndex][1].set_offsets(np.array(cspline_line[0],cspline_line[1]).T)
            else:
                self.lineplt[lineIndex][0].set_color('r')
                self.lineplt[lineIndex][1].set_xdata(np.array([showline[0,-1],showline[0,-1]]))
                self.lineplt[lineIndex][1].set_ydata(np.array([showline[0,-2],showline[0,-2]]))
                self.lineplt[lineIndex][1].set_color(temp_color)
                #self.lineplt[lineIndex][1].set_offsets(np.array([showline[0,[-1,-2]],showline[0,[-1,-2]]]))
            if self.line_selected>=0 and self.point_selected>=0:
                if type(self.lineplt[-1])==list:
                    self.lineplt.append(self.ax.scatter([self.lines[self.line_selected][self.point_selected][-1]],[self.lines[self.line_selected][self.point_selected][-2]],color='r',marker='x'))
                else:
                    self.lineplt[-1].set_offsets([self.lines[self.line_selected][self.point_selected][-1],self.lines[self.line_selected][self.point_selected][-2]])
                    self.lineplt[-1].set_visible(True)
            else:
                if type(self.lineplt[-1])!=list:
                    self.lineplt[-1].set_visible(False)
        self.fig.canvas.draw()
    def loadImageFrame(self):
        plt.clf()
        self.ax = self.fig.add_subplot(111)
        self.fig.subplots_adjust(bottom=(len(self.showIndex)+3)*0.04)
        showImage=getLastTwoDimArray(self.image.data,self.showIndex,color=self.color)
        if self.color:
            showImage[...,tuple(self.colorToggler)]=0
        showImage=np.maximum(0,np.minimum(255,(showImage*self.scaleVisual)**(10.**self.logVisual))).astype('uint8')
        self.main=self.ax.imshow(showImage,cmap=matplotlib.cm.gray, vmin=0, vmax=255)
        self.ax.set_aspect(self.image.dimlen[self.image.dim[-2-self.color]]/self.image.dimlen[self.image.dim[-1-self.color]])
        if self.contourImage is not None:
            showContour=getLastTwoDimArray(self.contourImage.data,self.showIndex,color=0)
            getlevels=np.arange(0.5+showContour.min(),showContour.max(),1)
            if len(getlevels)>0:
                self.contour=self.ax.contour(showContour,getlevels,linewidths=1.2)
            else:
                self.contour=None
        showpoints=getFramePts(self.points,self.showIndex)
        self.ptplt=self.ax.scatter(showpoints[:,-1],showpoints[:,-2],color='r',marker='x')
        if self.show_line:
            self.lineplt=[]
            for nline in range(len(self.lines)):
                temp_color='b'
                if nline==self.line_selected:
                    temp_color='r'
                if len(self.lines[nline])>0:
                    self.lineplt.append([self.ax.scatter(np.array(self.lines[nline])[:,-1],np.array(self.lines[nline])[:,-2],color='b',marker='x')])
                    if len(self.lines[nline])>1:
                        tck,temp = interpolate.splprep([np.array(self.lines[nline])[:,-1], np.array(self.lines[nline])[:,-2]], s=0,k=min(4,len(self.lines[nline]))-1)
                        cspline_line = interpolate.splev(np.arange(0, 1.01, 0.01), tck)
                        self.lineplt[-1]+=self.ax.plot(cspline_line[0].copy(),cspline_line[1].copy(),color=temp_color)
                    else:
                        self.lineplt[-1]+=self.ax.plot([self.lines[nline][0][-1],self.lines[nline][0][-1]],[self.lines[nline][0][-2],self.lines[nline][0][-2]],color=temp_color)
                        self.lineplt[-1][0].set_color('r')
                    if len(self.showIndex)>2:
                        if np.any(self.lines[nline][0][:(len(self.showIndex)-2)]!=self.showIndex[:-2]):
                            self.lineplt[-1][0].set_visible(False)
                            self.lineplt[-1][1].set_visible(False)
            if self.line_selected>=0 and self.point_selected>=0:
                self.lineplt.append(self.ax.scatter([self.lines[self.line_selected][self.point_selected][-1]],[self.lines[self.line_selected][self.point_selected][-2]],color='r',marker='x'))
        self.title=plt.title(self.addInstruct+dimToTitle(self.image.dim[:-2-self.color],self.showIndex[:-2]))
        plt.ylabel(self.image.dim[-2-self.color])
        plt.xlabel(self.image.dim[-1-self.color])

        '''set slider for image control'''
        axcolor = 'lightgoldenrodyellow'
        self.saveImageButton=Button(self.fig.add_axes([0.75, 0.02, 0.1, 0.05]), 'Save')
        self.saveImageButton.on_clicked(self.save_image)
        
        self.lineControl=[]
        if 'line' not in self.disable:
            self.lineControl.append(Button(self.fig.add_axes([0.05, 0.02, 0.1, 0.05]), 'Line'))
            self.lineControl[0].on_clicked(self.togger_line)
            self.lineControl.append(Button(self.fig.add_axes([0.2, 0.02, 0.1, 0.05]), 'new'))
            self.lineControl[1].on_clicked(self.new_line)
            self.lineControl.append(Button(self.fig.add_axes([0.35, 0.02, 0.1, 0.05]), 'edit...'))
            self.lineControl[2].on_clicked(self.edit_line)
            self.lineControl.append(Button(self.fig.add_axes([0.5, 0.02, 0.1, 0.05]), 'del...'))
            self.lineControl[3].on_clicked(self.del_line)
            
        self.axslide=[]
        self.sSlide=[]
        for n in range(len(self.showIndex)-2):
            self.axslide.append(self.fig.add_axes([0.1, self.sliderLoc+n*0.04, 0.65, 0.03], facecolor=axcolor))
            self.sSlide.append(Slider(self.axslide[-1], self.image.dim[n], 0, self.image.data.shape[n]-1, valinit=self.showIndex[n],valfmt="%i"))
            self.sSlide[-1].on_changed(self.sliderUpdate)
        self.axslide.append(self.fig.add_axes([0.1, self.sliderLoc+(len(self.showIndex)-2)*0.04, 0.65, 0.03], facecolor=axcolor))
        self.sSlide.append(Slider(self.axslide[-1], 'Iscale', 0.1, 10., valinit=self.scaleVisual,valstep=0.1))
        self.sSlide[-1].on_changed(self.sliderUpdate)
        self.axslide.append(self.fig.add_axes([0.1, self.sliderLoc+(len(self.showIndex)-1)*0.04, 0.65, 0.03], facecolor=axcolor))
        self.sSlide.append(Slider(self.axslide[-1], 'logPOW', -1., 1., valinit=self.logVisual,valstep=0.02))
        self.sSlide[-1].on_changed(self.sliderUpdate)
        
    def removeLastPoint(self):
        self.points=self.points[:-1,:]
        self.showNewPoints()
    def show(self):
        self.fig=plt.figure(1)
        self.sliderLoc=0.02
        self.connectionID=[]
        if 'click' not in self.disable or 'line' not in self.disable:
            self.connectionID.append(self.fig.canvas.mpl_connect('button_press_event', self.onclick))
        if 'line' not in self.disable:
            self.sliderLoc+=0.06
            self.connectionID.append(self.fig.canvas.mpl_connect('button_release_event', self.onrelease))
            self.connectionID.append(self.fig.canvas.mpl_connect('motion_notify_event', self.onmotion))
            
        self.connectionID.append(self.fig.canvas.mpl_connect('key_press_event',self.onKeypress))   
        self.enter=False
        self.loadImageFrame()
        plt.show()



'''
File: PointSpeckleProc.py
Description: Processes for ultrasound point speckles
History:
    Date    Programmer SAR# - Description
    ---------- ---------- ----------------------------
  Author: w.x.chan@gmail.com         04FEB2020           - Created
  
Requirements:
    numpy
    scipy
Known Bug:
    None
All rights reserved.
'''
_version='2.4.0'
import logging
logger = logging.getLogger(__name__)

import os
import numpy as np
from scipy.ndimage import shift
from scipy import signal
from scipy.ndimage.filters import gaussian_filter
from scipy.ndimage.filters import uniform_filter
from scipy.ndimage.measurements import variance
from scipy.stats import norm
import processFunc

#Optional dependancies

def gKern(std,kernlen=7,normalize=None):
    """Returns a Gaussian kernel array."""
    if isinstance(kernlen,(float,int)):
        kernlen=(kernlen*np.ones(len(std))).astype(int)
    gkern1d = []
    for n in range(len(std)):
        gkern1d.append(signal.gaussian(kernlen[n]*std, std=std[n]))
    gkern = np.zeros(gkern1d)
    gkern[:]=gkern1d[0].copy()
    gkern=gkern.transpose(np.roll(np.arange(len(std),dtype=int),-1))
    for n in range(1,len(std)):
        gkern[:]*=gkern1d[n]
        gkern=gkern.transpose(np.roll(np.arange(len(std),dtype=int),-1))
    if normalize:
        gkern*=normalize/gkern.max()
    return gkern2d

def getSpecklePoint(image,smoothingSize,prefilter=None):
    newImage=image.clone()
    extraDim=len(image.data.shape)-len(smoothingSize)
    if prefilter=='gaussian':
        tempArray=gaussian_filter(image.data, [*np.zeros(extraDim).astype(int),*smoothingSize],mode='constant')
        i=[]
        adjust=np.mgrid[tuple([slice(-1,2)]*len(smoothingSize))].reshape((2,-1)).T
        for adj in adjust:
            i.append(shift(tempArray,[*np.zeros(extraDim).astype(int),*adj],order=0))
        i=np.array(i)
        newImage.data[np.max(i,axis=0)>tempArray]=0
    else:
        i=[]
        sliceList=[]
        for n in range(len(smoothingSize)):
            sliceList.append(slice(-smoothingSize[n],smoothingSize[n]+1))
        adjust=np.mgrid[tuple(sliceList)].reshape((2,-1)).T
        for adj in adjust:
            i.append(shift(image.data,[*np.zeros(extraDim).astype(int),*adj],order=0))
        i=np.array(i)
        newImage.data[np.max(i,axis=0)>image.data]=0

    #remove duplicate
    nnzInd=np.transpose(np.nonzero(newImage.data))
    for checkInd in nnzInd:
        if newImage.data[tuple(checkInd)]>0:
            currentCheckIndex=checkInd.copy()
            for maxN in range(10):
                sliceList=list(currentCheckIndex[:extraDim])
                addind=np.zeros(len(smoothingSize),dtype=int)
                for nn in range(len(smoothingSize)):
                    sliceList.append(slice(max(0,currentCheckIndex[nn+extraDim]-int(np.ceil(smoothingSize[nn]))),min(newImage.data.shape[nn+extraDim],currentCheckIndex[nn+extraDim]+int(np.ceil(smoothingSize[nn]))+1)))
                    addind[nn]=max(0,currentCheckIndex[nn+extraDim]-int(np.ceil(smoothingSize[nn])))
                repeatedInd=np.transpose(np.nonzero(newImage.data[tuple(sliceList)]))
                if repeatedInd.shape[0]>1:
                    temp=np.around(np.mean(repeatedInd,axis=0)).astype(int)+addind
                    if np.all(temp==currentCheckIndex[extraDim:]):
                        break
                    else:
                        currentCheckIndex[extraDim:]=temp.copy()
                else:
                    break
            sliceList=list(currentCheckIndex[:extraDim])
            emptyslice=False
            for nn in range(len(smoothingSize)):
                sliceList.append(slice(max(0,currentCheckIndex[nn+extraDim]-int(np.ceil(smoothingSize[nn]))),min(newImage.data.shape[nn+extraDim],currentCheckIndex[nn+extraDim]+int(np.ceil(smoothingSize[nn]))+1)))
            temp_avg=newImage.data[tuple(sliceList)][newImage.data[tuple(sliceList)]>0].mean()
            newImage.data[tuple(sliceList)]=0
            newImage.data[tuple(currentCheckIndex)]=temp_avg
    return newImage
def reduceRepeat(image,checkSize,removeSingleton=0):
    reduce=image.data.shape[1]-1
    newImage=image.clone()
    for base in range(image.data.shape[1]):
        ii=[image.data[:,base].astype(bool)]
        for t in range(image.data.shape[1]):
            if t==base:
                continue
            i=[]
            for y in range(-checkSize[0],checkSize[0]+1):
                for x in range(-checkSize[1],checkSize[1]+1):
                            i.append(shift(image.data[:,t],[0,y,x],order=0))
            ii.append(np.any(np.array(i),axis=0))
        ii=np.array(ii).astype(int)
        temp_sum=np.maximum(np.sum(ii,axis=0),1).astype(float)
        if removeSingleton>0:
            newImage.data[:,base][temp_sum<=removeSingleton]=0
        newImage.data[:,base]*=(1-(temp_sum-1)/reduce)/temp_sum
    return newImage
        
def reduceNonRandom(image,sigmas,densityApprox=None,dimSigmaFactor=1.,average=False,truncate=0.5,dim='t',useCorrDet=False):
    oneNzeroArray=image.data.astype(bool).astype(float)
    normVal=(1./gaussian_filter(np.ones(np.ones(len(sigmas)).astype(int)), sigmas,mode='constant')).max()
    imgSigmaData=normVal*gaussian_filter(oneNzeroArray, [*np.zeros(len(image.data.shape)-len(sigmas)).astype(int),*sigmas],mode='constant')
    imgSigmaData=imgSigmaData.sum(axis=1)
    sigmaAvg=np.ones(len(sigmas))*np.mean(sigmas)
    if average:
        normValAvg=(1./gaussian_filter(np.ones(np.ones(len(sigmaAvg)).astype(int)), sigmaAvg,mode='constant',truncate=truncate)).max()
        imgSigmaAvgData=normValAvg*gaussian_filter(oneNzeroArray, [*np.zeros(len(image.data.shape)-len(sigmaAvg)).astype(int),*sigmaAvg],mode='constant',truncate=truncate)
        imgSigmaAvgData=imgSigmaAvgData.sum(axis=1)
    if densityApprox:
        density=densityApprox
    else:
        density=float(oneNzeroArray[oneNzeroArray>0].size)/oneNzeroArray.size
    dimInd=image.dim.index(dim)
    dimIndSlice=[slice(None)]*dimInd

    newImage=image.clone()
    
    if dimSigmaFactor!=0:
        mean=normVal*density*(image.data.shape[dimInd]-1)+1
        std=(mean-1)*abs(dimSigmaFactor)
        
        logger.info('Using mean and std of: {0:.3f} , {1:.3f}'.format(mean,std))
        nProb=norm(mean,std)
        normalize=1./nProb.pdf(mean)
        
        for t in range(image.data.shape[dimInd]):
            if average:
                newImage.data[(*dimIndSlice,t)][newImage.data[(*dimIndSlice,t)]>0]*=normalize*nProb.pdf(imgSigmaData[newImage.data[(*dimIndSlice,t)]>0])/imgSigmaAvgData[newImage.data[(*dimIndSlice,t)]>0]
            elif dimSigmaFactor<0:
                newImage.data[(*dimIndSlice,t)][newImage.data[(*dimIndSlice,t)]>0]=normalize*nProb.pdf(imgSigmaData[newImage.data[(*dimIndSlice,t)]>0])
            else:
                newImage.data[(*dimIndSlice,t)][newImage.data[(*dimIndSlice,t)]>0]*=normalize*nProb.pdf(imgSigmaData[newImage.data[(*dimIndSlice,t)]>0])
        if dimSigmaFactor<0:
            return newImage.data
    if useCorrDet:
        logger.warning('Warning, use only when sample size is large.')
        corrdet=[]
        posALL=[]
        maxCorrdet=0
        for t in range(image.data.shape[dimInd]):
            pos=np.array(np.nonzero(image.data[(*dimIndSlice,t)]>0))
            posALL.append(pos.copy())
            corrdet.append(np.zeros(pos.shape[1]))
            for n in range(pos.shape[1]):
                nearbyPosInd=np.nonzero(np.all(np.logical_and(pos[-2:]>=(pos[-2:,n]-np.ceil(sigmaAvg*2)).reshape((-1,1)),pos[-2:]<=(pos[-2:,n]+np.ceil(sigmaAvg*2)).reshape((-1,1))),axis=0))[0]
                if len(nearbyPosInd)>(3**2.):
                    temp=np.corrcoef(pos[:,nearbyPosInd])
                    if np.any(np.isnan(temp)):
                        temp=0
                    else:
                        temp=np.linalg.det(temp)
                    corrdet[-1][n]=temp
                elif np.any(pos[-2:,n]<np.ceil(sigmaAvg*2)) or np.any((pos[-2:,n]+np.ceil(sigmaAvg*2))>=image.data.shape[-2:]):
                        corrdet[-1][n]=-1
            if maxCorrdet<corrdet[-1].max():
                maxCorrdet=corrdet[-1].max()
        if maxCorrdet<0.5:
            logger.warning('Warning, determinant or correlation matrix has a maximum value of {0:.3e}'.format(maxCorrdet))
        for t in range(image.data.shape[dimInd]):
            corrdet[t]=corrdet[t]/maxCorrdet
            corrdet[t][corrdet[t]<0]=1.
            newImage.data[(*dimIndSlice,t)][tuple(posALL[t])]*=corrdet[t]
    return newImage
def applyThreshold(image,threshold):
    reduce=int(255/image.data.shape[1])
    newImage=image.clone()
    newImage.data[newImage.data<=(255-threshold*reduce)]=0
    return newImage

def singleOutFilter(speckleImageArray, sigma):
    threshold=speckleImageArray.max()*0.1
    newpoints=np.transpose(np.nonzero(speckleImageArray>threshold))
    val=speckleImageArray[np.nonzero(speckleImageArray>threshold)].copy()
    logger.info('Single out '+repr(len(newpoints))+' number of points')
    for n in range(len(newpoints)-1,-1,-1):
        toRemove=2
        for nn in range(-2,0,-1):
            if newpoints[n][nn]<(2*sigma[nn]) or newpoints[n][nn]>(speckleImageArray.shape[nn]-2*sigma[nn]-1):
                toRemove=0
        for nn in [[1,1],[1,-2],[-2,1],[-2,-2]]:
            if not(toRemove):
                break
            mincoord=newpoints[n][-2:]+np.array(nn)*sigma
            if np.any(np.logical_and(newpoints[:,-2:]>=mincoord,newpoints[:,-2:]<=(mincoord+sigma))):
                toRemove-=1
        if toRemove:
            newpoints=np.delete(newpoints,n,axis=0)
            val=np.delete(val,n,axis=0)
        elif toRemove==1:
            logger.debug('at least i have 1')
    logger.info('    to '+repr(len(newpoints))+' number of points')
    newpoints=np.around(newpoints).astype(int)
    newImageArray=np.zeros(speckleImageArray.shape)
    newImageArray[tuple(newpoints.T)]=val
    return newImageArray

def spreadSpeckle(image,spreadSize,overlay=False,overlayFunc=np.max,averageSigma=True,dim='t'):
    newImg=image.clone()
    percent99=np.percentile(newImg.data[newImg.data>0],99)
    if averageSigma:
        spreadSizeAvg=np.ones(len(spreadSize))*np.mean(spreadSize)
    else:
        spreadSizeAvg=spreadSize
    normVal=(1./gaussian_filter(np.ones(np.ones(len(spreadSizeAvg)).astype(int)), spreadSizeAvg,mode='constant')).max()
    if overlay:
        newImg.removeDim(dim)
        newImg.data=overlayFunc(image.data,axis=image.dim.index(dim))
    newImg.data=normVal*gaussian_filter(newImg.data, [*np.zeros(len(newImg.data.shape)-len(spreadSizeAvg)).astype(int),*spreadSizeAvg],mode='wrap')
    newImg.data*=255/percent99
    newImg.data=np.minimum(255,newImg.data)
    return newImg

def speckleTransform(speckleImageArray,transformFolder,fromTime,toTime=None,totalTimeSteps=None,Eulerian=True,highErrorDim=3):
    fromTime=int(fromTime)
    if type(toTime)!=type(None):
        toTime=int(toTime)
    pos=np.transpose(np.nonzero(speckleImageArray))[:,::-1]
    val=speckleImageArray[np.nonzero(speckleImageArray)].copy()
    if Eulerian:
        if totalTimeSteps:
            #Forward
            currentTime=fromTime
            if type(toTime)==type(None):
                temp_toTime=fromTime-1
            else:
                temp_toTime=toTime
            if temp_toTime<0:
                temp_toTime=totalTimeSteps-1
            posF=[pos.copy()]
            incr=1
            Fcount=0
            while currentTime!=temp_toTime:
                if currentTime+incr<totalTimeSteps:
                    nextTime=currentTime+incr
                else:
                    nextTime=0
                file=transformFolder+'/tstep'+str(currentTime)+'to'+str(nextTime)+'_0.txt'
                if not(os.path.isfile(file)):
                    logger.error('ERROR '+file+' does not exist')
                posF.append(processFunc.transform_img2img(posF[-1],file,savePath=transformFolder))
                currentTime=nextTime
                Fcount+=1
            #Backward
            currentTime=fromTime
            if type(toTime)==type(None):
                temp_toTime=fromTime+1
            else:
                temp_toTime=toTime
            if temp_toTime>=totalTimeSteps:
                temp_toTime=0
            posB=[pos.copy()]
            incr=-1
            Bcount=0
            while currentTime!=temp_toTime:
                if currentTime+incr>=0:
                    nextTime=currentTime+incr
                else:
                    nextTime=totalTimeSteps-1
                file=transformFolder+'/tstep'+str(currentTime)+'to'+str(nextTime)+'_0.txt'
                if not(os.path.isfile(file)):
                    logger.error('ERROR '+file+' does not exist')
                posB.append(processFunc.transform_img2img(posB[-1],file,savePath=transformFolder))
                currentTime=nextTime
                Bcount+=1
            if type(toTime)==type(None):
                Fratio=1./(1.+np.arange(totalTimeSteps)/np.arange(totalTimeSteps,0,-1))
                posF=np.roll(np.array(posF),fromTime,axis=0)
                Fratio=np.roll(Fratio,fromTime)
                posB=np.roll(np.array(posB)[::-1],fromTime+1,axis=0)
                newpos=Fratio.reshape((-1,1,1))*posF+(1-Fratio.reshape((-1,1,1)))*posB
            else:
                newpos=np.array([(Bcount*posF[-1]+Fcount*posB[-1])/(Fcount+Bcount)])
        else:
            currentTime=fromTime
            newpos=pos.copy()
            if currentTime>toTime:
                incr=-1
            elif currentTime<toTime:
                incr=1
            while currentTime!=toTime:
                file=transformFolder+'/tstep'+str(currentTime)+'to'+str(currentTime+incr)+'_0.txt'
                if not(os.path.isfile(file)):
                    logger.error('ERROR '+file+' does not exist')
                newpos=processFunc.transform_img2img(newpos,file,savePath=transformFolder)
                currentTime+=incr
            newpos=np.array([newpos])
    else:
        file=transformFolder+'/tstep'+str(fromTime)+'to'+str(toTime)+'_0.txt'
        if not(os.path.isfile(file)):
            logger.error('ERROR '+file+' does not exist')
        newpos=np.array([processFunc.transform_img2img(pos,file,savePath=transformFolder)])
    if type(toTime)==type(None) and Eulerian and totalTimeSteps:
        runN=totalTimeSteps
    else:
        runN=1
    newArray=[]
    newpos=np.around(newpos).astype(int)
    if highErrorDim and runN>1:
        error=Fratio*(1-Fratio)*totalTimeSteps
        if highErrorDim==True or highErrorDim==0:
            accScaling=1./np.maximum(1.,error)
        elif isinstance(highErrorDim,int):
            if highErrorDim<int((len(error)+1)/2):
                accScaling=1./np.maximum(1.,error/error[np.argmin(error)-highErrorDim])
            else:
                accScaling=np.ones(runN)
        else:
            accScaling=highErrorDim(error)
    else:
        accScaling=np.ones(runN)
    for n in range(runN):
        newArray.append(speckleImageArray.copy())
        get=np.all(np.logical_and(newpos[n]>=0,newpos[n]<np.array(speckleImageArray.shape)[::-1]),axis=-1)
        temppos=newpos[n][get]
        tempval=val[get]
        newArray[-1][:]=0
        newArray[-1][tuple(temppos[:,::-1].T)]=tempval.copy()*accScaling[n]
    if runN==1:
        newArray=newArray[0]
    return np.array(newArray)





'''
################################################
MIT License

Copyright (c) 2019 W. X. Chan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
################################################
File: __init__.py
Description: load all class for medImgProc
             Contains externally usable class
History:
    Date    Programmer SAR# - Description
    ---------- ---------- ----------------------------
  Author: w.x.chan@gmail.com         12JAN2019           - Created
  Author: w.x.chan@gmail.com         12JAN2019           - v1.3.2
                                                            -processFunc v1.3.0
                                                            -Image v1.0.0
                                                            -GUI v1.0.0
  Author: w.x.chan@gmail.com         12JAN2019           - v1.4.2
                                                            -processFunc v1.3.0
                                                            -Image v1.4.2
                                                            -GUI v1.4.0
  Author: w.x.chan@gmail.com         08OCT2019           - v1.5.2
                                                            -processFunc v1.3.0
                                                            -Image v1.4.2
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         08OCT2019           - v1.5.4
                                                            -processFunc v1.3.0
                                                            -Image v1.5.4
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         15OCT2019           - v1.5.5
                                                            -processFunc v1.3.0
                                                            -Image v1.5.5
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         15OCT2019           - v1.6.1
                                                                -added combine grescale image to color
                                                            -processFunc v1.3.0
                                                            -Image v1.5.5
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         29OCT2019           - v1.6.3
                                                                -added point based combine to function combine
                                                            -processFunc v1.3.0
                                                            -Image v1.6.2
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         29OCT2019           - v1.6.4
                                                            -processFunc v1.3.0
                                                            -Image v1.6.4
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         29OCT2019           - v1.6.5
                                                            -processFunc v1.3.0
                                                            -Image v1.6.5
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         30OCT2019           - v1.6.6
                                                            -processFunc v1.6.6
                                                            -Image v1.6.5
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         30OCT2019           - v1.7.0
                                                            -processFunc v1.7.0
                                                            -Image v1.6.5
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         31OCT2019           - v1.7.3
                                                            -processFunc v1.7.3
                                                            -Image v1.6.5
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         08NOV2019           - v1.7.6
                                                            -processFunc v1.7.3
                                                            -Image v1.7.6
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         08NOV2019           - v1.8.0
                                                            -processFunc v1.7.3
                                                            -Image v1.8.0
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         13NOV2019           - v1.8.5
                                                            -processFunc v1.8.5
                                                            -Image v1.8.0
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         13NOV2019           - v1.9.1
                                                            -processFunc v1.9.1
                                                            -Image v1.8.0
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         13NOV2019           - v2.0.0
                                                            -processFunc v2.0.0
                                                            -Image v1.8.0
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         18NOV2019           - v2.1.7
                                                            -processFunc v2.1.5
                                                            -Image v1.8.0
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         02JAN2020           - v2.2.8
                                                            -processFunc v2.2.8
                                                            -Image v1.8.0
                                                            -GUI v1.5.2
  Author: w.x.chan@gmail.com         10JAN2020           - v2.3.11
                                                            -processFunc v2.2.8
                                                            -Image v1.8.0
                                                            -GUI v2.3.10
  Author: w.x.chan@gmail.com         15JAN2020           - v2.3.13
                                                            -allow loadStack to add new dimension and stack at different axis
                                                            -processFunc v2.2.8
                                                            -Image v2.3.13
                                                            -GUI v2.3.10
  Author: w.x.chan@gmail.com         21JAN2020           - v2.3.14
                                                            -processFunc v2.3.14
                                                            -Image v2.3.13
                                                            -GUI v2.3.10
  Author: w.x.chan@gmail.com         21JAN2020           - v2.4.0
                                                            -processFunc v2.4.0
                                                            -Image v2.3.13
                                                            -GUI v2.3.10
                                                            -pointSpeckleProc v2.4.0
                                                            -Snake v2.4.0
  Author: w.x.chan@gmail.com         21JAN2020           - v2.4.3
                                                            -processFunc v2.4.0
                                                            -Image v2.4.1
                                                            -GUI v2.3.10
                                                            -pointSpeckleProc v2.4.0
  Author: w.x.chan@gmail.com         26FEB2020           - v2.4.6
                                                            -processFunc v2.4.6
                                                            -Image v2.4.1
                                                            -GUI v2.3.10
                                                            -pointSpeckleProc v2.4.0
  Author: w.x.chan@gmail.com         28FEB2020           - v2.5.0
                                                            -processFunc v2.5.0
                                                            -Image v2.4.1
                                                            -GUI v2.3.10
                                                            -pointSpeckleProc v2.4.0
  Author: w.x.chan@gmail.com         05Mar2020           - v2.5.2
                                                            -processFunc v2.5.1
                                                            -Image v2.4.1
                                                            -GUI v2.3.10
                                                            -pointSpeckleProc v2.4.0
  Author: w.x.chan@gmail.com         17Mar2020           - v2.5.7
                                                            -processFunc v2.5.7
                                                            -Image v2.4.1
                                                            -GUI v2.3.10
                                                            -pointSpeckleProc v2.4.0
  Author: w.x.chan@gmail.com         24Mar2020           - v2.6.13 - in combine, transfer boolean image to contour 
                                                            -processFunc v2.5.7
                                                            -Image v2.6.10
                                                            -GUI v2.6.13
                                                            -pointSpeckleProc v2.4.0
  Author: w.x.chan@gmail.com         25Mar2020           - v2.6.15
                                                            -processFunc v2.5.7
                                                            -Image v2.6.15
                                                            -GUI v2.6.13
                                                            -pointSpeckleProc v2.4.0


Requirements:
    numpy.py
    matplotlib.py
    imageio.py

Known Bug:
    HSV color format not supported
All rights reserved.
'''
import logging
_version='2.6.15'
logger = logging.getLogger('medImgProc v'+_version)
logger.info('medImgProc version '+_version)


import numpy as np
import os
import sys
sys.path.insert(0, os.path.dirname(os.path.realpath(__file__)))
import matplotlib
if os.environ.get('DISPLAY','')=='' and os.name!='nt':
    logger.warning('no display found. Using non-interactive Agg backend')
    matplotlib.use('Agg')
import pickle
try:
    import scipy.io as sio
except:
    pass

import image

'''
External Functions
'''
def imread(imageFile,dimension=None,fileFormat='',crop=None,module=''):
    newImage=image.image(imageFile,dimension=dimension,fileFormat=fileFormat,crop=crop,module=module)
    return newImage
def imwrite(imageClass,filePath,axes=['y','x'],imageFormat='png',dimRange={},fps=3,color=0):
    imageClass.imwrite2D(filePath,axes=axes,imageFormat=imageFormat,dimRange=dimRange,fps=fps,color=color)
def stackImage(imageList,newDim):
    newImage=imageList[0].clone()
    newImage.addDim(newDim)
    newImage.insertnewImageList(imageList[1:],newDim)
    return newImage
def apply(imageClass,func,axes=['y','x'],dimSlice={},funcArgs=()):#use slice(a,b) for otherDimLoc 
    newImage=image.applyFunc(imageClass,func,axes,dimSlice,funcArgs)
    return newImage
def arrange(imageClass,newDim,arrangeFront=True):
    newImage=imageClass.clone()
    transposeIndex,currentDim=image.arrangeDim(newImage.dim[:],newDim,arrangeFront)
    newImage.data=newImage.data.transpose(transposeIndex)
    newImage.dim=currentDim[:]
    return newImage
def stretch(imageClass,stretchDim,scheme=image.DEFAULT_INTERPOLATION_SCHEME):
    newImage=imageClass.clone()
    newImage.stretch(stretchDim,scheme=scheme)
    return newImage
def save(imageClass,filePath):
    imageClass.save(filePath)
def load(filePath):
    try:
        with open(filePath, 'rb') as input:
            outObj = pickle.load(input)
    except:
        outObj=imread(filePath)
    return outObj
def loadStack(imageFileFormat,dimension=None,n=0,maxskip=0):
    getFunc=None
    if isinstance(n,(list,tuple)):
        if len(n)==0:
            n=0
            n_new=0
        elif len(n)==1:
            n=n[0]
            n_new=n
        else:
            n_new=list(n)
            n=n_new.pop(0)
    else:
        n_new=n
    if isinstance(dimension,(list,tuple)):
        dimension_new=[]
        if len(dimension)==0:
            dimension=None
        else:
            dimension_new=list(dimension)
            dimension=dimension_new.pop(0)
        if len(dimension_new)>0:
            getFunc=lambda x_imageFileFormat : loadStack(x_imageFileFormat,dimension=dimension_new,n=n_new,maxskip=maxskip)
            newImage=getFunc(imageFileFormat.format(n))
    if type(getFunc)==type(None):
        try:
            newImage=load(imageFileFormat.format(n))
            getFunc=load
        except:
            newImage=imread(imageFileFormat.format(n))
            getFunc=imread
    if type(dimension)==type(None):
        stackaxis=0
    elif isinstance(dimension,int):
        stackaxis=dimension
    elif dimension in newImage.dim:
        stackaxis=newImage.dim.index(dimension)
    else:
        newImage.addDim(dimension)
        stackaxis=0
    n+=1
    skip=0
    while True:
        try:
            nextImg=getFunc(imageFileFormat.format(n))
            if dimension not in nextImg.dim:
                nextImg.data=nextImg.data.reshape((1,*nextImg.data.shape))
            newImage.data=np.concatenate((newImage.data,nextImg.data.copy()),axis=stackaxis)
            skip=0
        except:
            skip+=1
            if skip>maxskip:
                break
            print(n)
        n+=1
    return newImage
def loadmat(fileName,arrayName='',dim=[],dimlen={},dtype=None):
    newImage=image.image()
    try:
        matVariables=sio.loadmat(fileName)
        if not(arrayName):
            for key in matVariables:
                if type(matVariables[key])==np.ndarray:
                    newImage.data=matVariables[key]
                    break
            else:
                print('Error Loading matlab file.')
                return
    except NotImplementedError:
        import h5py
        matVariables = h5py.File(fileName)
        if not(arrayName):
            for key in matVariables:
                print(type(matVariables[key].value))
                if type(matVariables[key].value)==np.ndarray:
                    newImage.data=matVariables[key].value
                    break
            else:
                print('Error Loading matlab file.')
                return
    if len(dim)!=len(newImage.data.shape):
        newImage.dim=image.DEFAULT_SEGMENTATION_DIMENSION[-len(newImage.data.shape):]
    else:
        newImage.dim=dim[:]
    newImage.dimlen=dict(dimlen)
    if len(newImage.dim)!=len(newImage.dimlen):
        for dimension in newImage.dim:
            if dimension not in newImage.dimlen:
                newImage.dimlen[dimension]=1.
    if dtype is None:
        newImage.dtype=newImage.data.dtype
    else:
        newImage.data=newImage.data.astype(dtype)
        newImage.dtype=dtype
    return newImage
def show(imageClass):
    imageClass.show()

def combine(img1,img2,img3=None,point=False):
    #img1 is main image
    img=img1.clone()
    if img1.data.dtype==bool:
        img.changeBooleanToContour()
    img.changeColorFormat()
    if type(img3)!=type(None):
        if img2.data.dtype==bool:
            img2=img2.clone()
            img2.changeBooleanToContour()
        if img3.data.dtype==bool:
            img3=img3.clone()
            img3.changeBooleanToContour()
        img.data[...,1]=img2.data.copy()
        img.data[...,2]=img3.data.copy()
    elif point:
        img.data[...,0][img2.data>0]=img2.data[img2.data>0]
        img.data[...,1][img2.data>0]=255
        img.data[...,2][img2.data>0]=0
    else:
        if img2.data.dtype==bool:
            img2=img2.clone()
            img2.changeBooleanToContour()
        img.data[...,0]=np.maximum(img2.data,img1.data)
    return img
